//mixer.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2012-2013
 *
 *  This file is part of SavannahIce,
 *  a protocol converter between RoarAudio and Icecast.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  RoarAudio is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "savannahice.h"
#include <shout/shout.h>

#define MAX_STREAMS 4

#define LIBSHOUT_THRESHOLD (1024*8)

#define __ckself  if ( self == NULL ) { roar_err_set(ROAR_ERROR_FAULT); return -1; }


struct savannahice_mixer {
 size_t refc;
 shout_t * shout;
 savannahice_stream_t streams[MAX_STREAMS];
 int shout_format;
 savannahice_stream_t mixerstream;
 int is_connected;
};

static void __shouterr2roarerr(savannahice_mixer_t self) {
 int err;

 switch (shout_get_errno(self->shout)) {
  default: err = ROAR_ERROR_UNKNOWN; break;
 }

 roar_err_set(err);
}

static void __next_stream(savannahice_mixer_t self) {
 size_t i;

 ROAR_DBG("__next_stream(self=%p) = ?", self);

 ROAR_DBG("__next_stream(self=%p): self->streams[0] = %p, self->streams[1] = %p", self, self->streams[0], self->streams[1]);
 savannahice_stream_unref(self->streams[0]);

 for (i = 1; i < MAX_STREAMS; i++) {
  self->streams[i-1] = self->streams[i];
 }

 self->streams[MAX_STREAMS-1] = NULL;
}

static int __get_shout_format_from_codec(uint32_t codec) {
  switch (codec) {
   case ROAR_CODEC_OGG_VORBIS:
   case ROAR_CODEC_OGG_SPEEX:
   case ROAR_CODEC_OGG_FLAC:
   case ROAR_CODEC_OGG_GENERAL:
   case ROAR_CODEC_OGG_CELT:
   case ROAR_CODEC_OGG:
#ifdef ROAR_CODEC_OGG_OPUS
   case ROAR_CODEC_OGG_OPUS:
#endif
     return SHOUT_FORMAT_OGG;
    break;
   default:
     roar_err_set(ROAR_ERROR_NOTSUP); // ROAR_ERROR_NSTYPE?
     return -1;
    break;
  }
}

static int __get_shout_format_from_stream(savannahice_stream_t stream) {
 uint32_t codec;

 codec = savannahice_stream_get_codec(stream);

 return __get_shout_format_from_codec(codec);
}

static int __get_shout_format(savannahice_mixer_t self) {
 int ret = -1;

 while (ret == -1) {
  if ( self->streams[0] == NULL ) {
   roar_err_set(ROAR_ERROR_NODATA);
   return -1;
  }

  ret = __get_shout_format_from_stream(self->streams[0]);
  if ( ret == -1 )
   __next_stream(self);
 }

 return ret;
}

savannahice_mixer_t savannahice_mixer_new(void) {
 savannahice_mixer_t ret = roar_mm_malloc(sizeof(struct savannahice_mixer));
 size_t i;
 int err;

 if ( ret == NULL )
  return NULL;

 memset(ret, 0, sizeof(*ret));

 ret->refc = 1;
 ret->is_connected = 0;
 ret->shout_format = -1;

 for (i = 0; i < MAX_STREAMS; i++)
  ret->streams[i] = NULL;

 ret->mixerstream = savannahice_stream_new(0, ROAR_DIR_MIXING, 0 /* FIXME */);
 if ( ret->mixerstream == NULL ) {
  err = roar_error;
  roar_mm_free(ret);
  roar_err_set(err);
  return NULL;
 }

 ret->shout = shout_new();
 if ( ret->shout == NULL ) {
  roar_mm_free(ret);
  roar_err_set(ROAR_ERROR_IO);
  return NULL;
 }

 return ret;
}

savannahice_mixer_t savannahice_mixer_new_simple(const char * node, uint16_t port, const char * mount, const char * user, const char * password) {
 savannahice_mixer_t ret = savannahice_mixer_new();
 int err;

 if ( ret == NULL )
  return NULL;

 if ( savannahice_mixer_set_node(ret, node, port)     != 0 ||
      savannahice_mixer_set_mount(ret, mount)         != 0 ||
      savannahice_mixer_set_auth(ret, user, password) != 0 ) {
  err = roar_error;
  savannahice_mixer_unref(ret);
  roar_err_set(err);
  return NULL;
 }

 return ret;
}

int savannahice_mixer_set_node(savannahice_mixer_t self, const char * node, uint16_t port) {
 __ckself

 if ( node != NULL )
  shout_set_host(self->shout, node);

 if ( port != 0 )
  shout_set_port(self->shout, port);

 return 0;
}

int savannahice_mixer_set_mount(savannahice_mixer_t self, const char * mount) {
 __ckself

 if ( mount == NULL )
  mount = "/SavannahIce.ogg";

 shout_set_mount(self->shout, mount);
 return 0;
}

int savannahice_mixer_set_auth(savannahice_mixer_t self, const char * user, const char * password) {
 __ckself

 if ( password == NULL )
  password = "hackme";

 if ( user != NULL )
  shout_set_user(self->shout, user);

 shout_set_password(self->shout, password);
 return 0;
}

int savannahice_mixer_ref(savannahice_mixer_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc++;

 return 0;
}

int savannahice_mixer_unref(savannahice_mixer_t self) {
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc--;

 if (self->refc == 0) {
  for (i = 0; i < MAX_STREAMS; i++)
   if ( self->streams[i] != NULL )
    savannahice_stream_unref(self->streams[i]);

  if ( self->shout != NULL )
   shout_free(self->shout);

  savannahice_stream_unref(self->mixerstream);
  roar_mm_free(self);
 }

 return 0;
}

int savannahice_mixer_connect(savannahice_mixer_t self) {
 int ret;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( self->shout == NULL ) {
  roar_err_set(ROAR_ERROR_NODEV);
  return -1;
 }

 if ( self->is_connected ) {
  roar_err_set(ROAR_ERROR_ALREADY);
  return -1;
 }

 self->shout_format = __get_shout_format(self);
 if ( self->shout_format == -1 )
  return -1;


 shout_set_format(self->shout, self->shout_format);
 shout_set_nonblocking(self->shout, 1);

 ret = shout_open(self->shout);
 if ( ret != SHOUTERR_SUCCESS && ret != SHOUTERR_NOCONNECT && ret != SHOUTERR_BUSY ) {
  __shouterr2roarerr(self);
  return -1;
 }

 self->is_connected = 1;
 return 0;
}

int savannahice_mixer_handle(savannahice_mixer_t self) {
 struct roar_buffer ** buffer;
 ssize_t shoutqueue;
 int format;
 size_t len;
 size_t writesize;
 void * buf;
 int ret;

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( self->streams[0] == NULL ) {
  roar_err_set(ROAR_ERROR_NODATA);
  return -1;
 }

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( self->shout == NULL ) {
  roar_err_set(ROAR_ERROR_NODEV);
  return -1;
 }

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( !self->is_connected )
  if ( savannahice_mixer_connect(self) == -1 )
   return -1;

 if ( shout_get_connected(self->shout) != SHOUTERR_CONNECTED )
  return -1;

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 shoutqueue = shout_queuelen(self->shout);
 if ( shoutqueue < 0 ) {
  __shouterr2roarerr(self);
  return -1;
 }

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( shoutqueue > LIBSHOUT_THRESHOLD )
  return 0;

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 buffer = savannahice_stream_get_buffer(self->streams[0]);

 ROAR_DBG("savannahice_mixer_handle(self=%p): buffer=%p", self, buffer);

 if ( buffer == NULL || *buffer == NULL ) {
  ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);
  if ( savannahice_stream_get_alone(self->streams[0]) ) {
   ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);
   __next_stream(self); // move to next stream, will work on it in next iteration.
   ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

   // skip streams of wrong format.
   format = __get_shout_format(self);
   while (format != self->shout_format) {
    if ( self->streams[0] == NULL ) {
     roar_err_set(ROAR_ERROR_NODATA);
     return -1;
    }
    ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);
    __next_stream(self);
    format = __get_shout_format(self);
   }
  }

  roar_err_set(ROAR_ERROR_NODATA);
  return -1;
 }

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 writesize = LIBSHOUT_THRESHOLD - (size_t)shoutqueue;

 if ( roar_buffer_get_len(*buffer, &len) == -1 )
  return -1;

 if ( roar_buffer_get_data(*buffer, &buf) == -1 )
  return -1;

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( len < writesize )
  writesize = len;

 ret = shout_send(self->shout, buf, writesize);
// shout_sync(self->shout); // FIXME: We shoudn't do this here.
// shout_send(self->shout, NULL, 0); // trigger flush.

 // do something the the return value later. the bytes also got eaten up in case of error.

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( writesize == len ) {
  ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);
  if ( roar_buffer_next(buffer) == -1 ) {
   return -1;
  }
 } else {
  ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);
  if ( roar_buffer_set_offset(*buffer, writesize) == -1 ) {
   return -1;
  }
 }

 ROAR_DBG("savannahice_mixer_handle(self=%p) = ?", self);

 if ( ret == SHOUTERR_SOCKET ) {
  // try to recover from timeouts.
  shout_close(self->shout);
  self->is_connected = 0;
  __next_stream(self);
 }

 if ( ret != SHOUTERR_SUCCESS ) {
  __shouterr2roarerr(self);
  return -1;
 }

 ROAR_DBG("savannahice_mixer_handle(self=%p) = 0", self);
 return 0;
}

int savannahice_mixer_attach_stream(savannahice_mixer_t self, savannahice_stream_t stream) {
 size_t i;
 int format;

 if ( self == NULL || stream == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 format = __get_shout_format_from_stream(stream);

 if ( format == -1 )
  return -1;

 if ( self->shout_format != -1 && self->shout_format != format )
  return -1;

 for (i = 0; i < MAX_STREAMS; i++) {
  if ( self->streams[i] == NULL ) {
   if ( savannahice_stream_ref(stream) == -1 )
    return -1;
   self->streams[i] = stream;
   return 0;
  }
 }

 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}

int savannahice_mixer_set_codec(savannahice_mixer_t self, uint32_t codec) {
 int format;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 format = __get_shout_format_from_codec(codec);

 if ( format == -1 )
  return -1;

 if ( self->shout_format != -1 && self->shout_format != format ) {
  roar_err_set(ROAR_ERROR_BUSY);
  return -1;
 }

 self->shout_format = format;

 return 0;
}

int_least32_t savannahice_mixer_delay(savannahice_mixer_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return (int_least32_t)-1;
 }

 if ( self->shout == NULL ) {
  roar_err_set(ROAR_ERROR_NOENT);
  return (int_least32_t)-1;
 }

 return (int_least32_t)shout_delay(self->shout) * (int_least32_t)1000;
}

savannahice_stream_t savannahice_mixer_stream(savannahice_mixer_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 if ( savannahice_stream_ref(self->mixerstream) == -1 )
  return NULL;

 return self->mixerstream;
}

//ll
