//network.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2012-2013
 *
 *  This file is part of SavannahIce,
 *  a protocol converter between RoarAudio and Icecast.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  RoarAudio is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "savannahice.h"

#define MAX_LISTENS 4
#define MAX_PROTOS  4

enum status {
 STATUS_UNUSED = 0,
 STATUS_USED   = 1
};

struct client {
 enum status status;
 struct roar_vio_calls * vio;
 int proto;
 struct roar_buffer * obuf;
 void * userdata;
};

struct listen {
 enum status status;
 struct roar_vio_calls vio;
 int proto;
};

struct savannahice_network {
 size_t refc;
 size_t clients_len;
 struct client * clients;
 struct listen listens[MAX_LISTENS];
 savannahice_protocol_t protos[MAX_PROTOS];
 struct roar_dl_fnreg callback;
};

static void __delete_client(savannahice_network_t self, size_t id);

// this function *tries* to ensure space for at least *two* more client.
static void __ensure_client_space(savannahice_network_t self) {
 size_t i;
 size_t newsize;
 struct client * p;
 size_t free_clients = 0;

 for (i = 0; i < self->clients_len; i++)
  if ( self->clients[i].status == STATUS_UNUSED )
   free_clients++;

 if ( free_clients >= 2 )
  return;

 newsize = self->clients_len + 8;
 p = roar_mm_realloc(self->clients, sizeof(struct client)*newsize);
 if ( p == NULL )
  return;

 self->clients = p;

 for (i = self->clients_len; i < newsize; i++)
  p[i].status = STATUS_UNUSED;

 self->clients_len = newsize;
}

static int plugin_callback(enum roar_dl_fnreg_action action, int fn, int subtype, const void * object, size_t objectlen, int version, int options, void * userdata, struct roar_dl_lhandle * lhandle) {
 const struct roar_dl_proto * impl = object;
 savannahice_network_t self = userdata;
 size_t i;

 (void)fn, (void)subtype, (void)version, (void)options;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( objectlen != ROAR_DL_PROTO_SIZE ) {
  ROAR_WARN("plugin_callback(*): Library %p tries to register protocol with bad object length.", lhandle);
  roar_err_set(ROAR_ERROR_BADLIB);
  return -1;
 }

 switch (action)  {
  case ROAR_DL_FNREG:
    for (i = 0; i < MAX_PROTOS; i++) {
     if ( !self->protos[i].used ) {
      memset(&(self->protos[i]), 0, sizeof(self->protos[i]));
      self->protos[i].used = 1;
      self->protos[i].para = NULL;
      self->protos[i].paralen = -1;
      self->protos[i].pluginpara = NULL;
      self->protos[i].lhandle = lhandle;
      self->protos[i].impl = impl;
      return 0;
     }
    }
   break;
  case ROAR_DL_FNUNREG:
    for (i = 0; i < self->clients_len; i++) {
     if ( self->clients[i].status != STATUS_UNUSED && self->clients[i].proto == impl->proto ) {
      __delete_client(self, i);
     }
    }
    for (i = 0; i < MAX_PROTOS; i++) {
     if ( self->protos[i].used && self->protos[i].lhandle == lhandle ) {
      self->protos[i].used = 0;
     }
    }
   break;
 }

 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}


// this function returns true if there is free client space or false if not or error.
static int __check_for_client_space(savannahice_network_t self) {
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return 0;
 }

 for (i = 0; i < self->clients_len; i++)
  if ( self->clients[i].status == STATUS_UNUSED )
   return 1;

 roar_err_set(ROAR_ERROR_NOSPC);
 return 0;
}

static const savannahice_protocol_t * __lookup_proto(savannahice_network_t self, int proto) {
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return NULL;
 }

 for (i = 0; i < MAX_PROTOS; i++)
  if ( self->protos[i].used )
   if ( self->protos[i].impl->proto == proto )
    return &(self->protos[i]);

 roar_err_set(ROAR_ERROR_NOENT);
 return NULL;
}

static void __delete_client(savannahice_network_t self, size_t id) {
 const savannahice_protocol_t * proto;

 ROAR_DBG("__delete_client(self=%p, id=%zu) = ?", self, id);

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return;
 }

 if ( id > self->clients_len ) {
  roar_err_set(ROAR_ERROR_INVAL);
  return;
 }

 if ( self->clients[id].status != STATUS_USED ) {
  roar_err_set(ROAR_ERROR_NOENT);
  return;
 }

 proto = __lookup_proto(self, self->clients[id].proto);
 if ( proto != NULL ) {
  if ( proto->impl->unset_proto != NULL ) {
   if ( proto->lhandle != NULL )
    roar_dl_context_restore(proto->lhandle);
   proto->impl->unset_proto(id, self->clients[id].vio, &(self->clients[id].obuf), &(self->clients[id].userdata), proto->para, proto->paralen, proto->pluginpara);
   if ( proto->lhandle != NULL )
    roar_dl_context_store(proto->lhandle);
  }
 }

 roar_vio_close(self->clients[id].vio);

 if ( self->clients[id].obuf != NULL )
  roar_buffer_free(self->clients[id].obuf);

 if ( self->clients[id].userdata != NULL )
  roar_mm_free(self->clients[id].userdata);

 self->clients[id].status = STATUS_UNUSED;

 roar_err_set(ROAR_ERROR_NONE);
}

static void __flush_client(savannahice_network_t self, size_t id, struct roar_vio_calls * vio, struct roar_buffer ** buffer) {
 size_t len;
 ssize_t ret;
 void * buf;

 if ( roar_buffer_get_len(*buffer, &len) == -1 )
  return;

 if ( roar_buffer_get_data(*buffer, &buf) == -1 )
  return;

 ret = roar_vio_write(vio, buf, len);

 if ( ret < 1 ) {
  __delete_client(self, id);
  return;
 }

 if ( ret == (ssize_t)len ) {
  if ( roar_buffer_next(buffer) == -1 ) {
   __delete_client(self, id);
   return;
  }
 } else {
  if ( roar_buffer_set_offset(*buffer, ret) == -1 ) {
   __delete_client(self, id);
   return;
  }
 }

 roar_err_set(ROAR_ERROR_NONE);
}

savannahice_network_t savannahice_network_new(void) {
 savannahice_network_t ret = roar_mm_malloc(sizeof(struct savannahice_network));
 size_t i;
 int err_clients;

 if ( ret == NULL )
  return NULL;

 memset(ret, 0, sizeof(*ret));

 ret->refc = 1;
 ret->clients_len = 0;

 __ensure_client_space(ret);
 err_clients = roar_error;

 for (i = 0; i < MAX_LISTENS; i++)
  ret->listens[i].status = STATUS_UNUSED;
 for (i = 0; i < MAX_PROTOS; i++)
  ret->protos[i].used = 0;

 ret->callback.fn = ROAR_DL_FN_PROTO;
 ret->callback.subtype = ROAR_DL_PROTO_SUBTYPE;
 ret->callback.version = ROAR_DL_PROTO_VERSION;
 ret->callback.callback = plugin_callback;
 ret->callback.userdata = ret;

 ROAR_DL_RFNREG(ROAR_DL_HANDLE_APPLICATION, ret->callback);

 if ( ret->clients_len == 0 ) {
  savannahice_network_unref(ret);
  roar_err_set(err_clients);
  return NULL;
 }

 return ret;
}

int savannahice_network_ref(savannahice_network_t self) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc++;

 return 0;
}

int savannahice_network_unref(savannahice_network_t self) {
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 self->refc--;

 if ( self->refc == 0 ) {
  savannahice_network_del_listen_all(self);

  for (i = 0; i < self->clients_len; i++)
   __delete_client(self, i);

  if ( self->clients != NULL )
   roar_mm_free(self->clients);

  roar_mm_free(self);
 }

 return 0;
}

int savannahice_network_del_listen_all(savannahice_network_t self) {
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 for (i = 0; i < MAX_LISTENS; i++) {
  if ( self->listens[i].status == STATUS_USED ) {
   roar_vio_close(&(self->listens[i].vio));
   self->listens[i].status = STATUS_UNUSED;
  }
 }

 return 0;
}

int savannahice_network_add_listen_address(savannahice_network_t self, int type, const char * address, int port, int proto) {
 size_t i;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( type == ROAR_SOCKET_TYPE_TCP && port == 0 )
  port = ROAR_DEFAULT_PORT;

 if ( address == NULL ) {
#ifdef DEFAULT_SERVER_ADDRESS
  address = DEFAULT_SERVER_ADDRESS;
#else
  switch (type) {
   case ROAR_SOCKET_TYPE_TCP:
     address = ROAR_DEFAULT_HOST;
    break;
   case ROAR_SOCKET_TYPE_DECNET:
     address = ROAR_DEFAULT_LISTEN_OBJECT;
    break;
   case ROAR_SOCKET_TYPE_UNIX:
   case ROAR_SOCKET_TYPE_UNKNOWN:
     address = DEFAULT_UNIX_ADDRESS;
     type = ROAR_SOCKET_TYPE_UNIX;
    break;
  }
#endif
 }

 if ( address == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 for (i = 0; i < MAX_LISTENS; i++) {
  if ( self->listens[i].status == STATUS_UNUSED ) {
   if ( roar_vio_open_socket_listen(&(self->listens[i].vio), type, address, port) == -1 ) {
    if ( roar_error == ROAR_ERROR_ADDRINUSE && (type == ROAR_SOCKET_TYPE_UNIX || address[0] == '/') ) {
     unlink(address);
     if ( roar_vio_open_socket_listen(&(self->listens[i].vio), type, address, port) == -1 )
      return -1;
    } else {
     return -1;
    }
   }

   self->listens[i].status = STATUS_USED;
   self->listens[i].proto = proto;
   return 0;
  }
 }

 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}

int savannahice_network_add_client_fh(savannahice_network_t self, int fh, int proto) {
 const savannahice_protocol_t * protoobj;
 size_t i;
 int err;

 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( fh < 0 ) {
  roar_err_set(ROAR_ERROR_BADFH);
  return -1;
 }

 protoobj = __lookup_proto(self, proto);
 if ( protoobj == NULL )
  return -1;

 __ensure_client_space(self);

 for (i = 0; i < self->clients_len; i++) {
  if ( self->clients[i].status == STATUS_UNUSED ) {
   self->clients[i].vio = roar_mm_malloc(sizeof(*(self->clients[i].vio)));
   if ( self->clients[i].vio == NULL )
    return -1;
   if ( roar_vio_open_fh_socket(self->clients[i].vio, fh) == -1 ) {
    err = roar_error;
    roar_mm_free(self->clients[i].vio);
    roar_err_set(err);
    return -1;
   }
   self->clients[i].vio->flags |= ROAR_VIO_FLAGS_FREESELF;

   self->clients[i].status   = STATUS_USED;
   self->clients[i].proto    = proto;
   self->clients[i].obuf     = NULL;
   self->clients[i].userdata = NULL;

   if ( protoobj->impl->set_proto != NULL ) {
    if ( protoobj->lhandle != NULL )
     roar_dl_context_restore(protoobj->lhandle);
    protoobj->impl->set_proto(i, self->clients[i].vio, &(self->clients[i].obuf), &(self->clients[i].userdata), protoobj->para, protoobj->paralen, protoobj->pluginpara);
    if ( protoobj->lhandle != NULL )
     roar_dl_context_store(protoobj->lhandle);
   }
   return i;
  }
 }

 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}

static int savannahice_network_add_client_accept(savannahice_network_t self, struct roar_vio_calls * listen, int proto) {
 const savannahice_protocol_t * protoobj;
 struct roar_vio_calls vio;
 size_t i;
 int err;

 ROAR_DBG("savannahice_network_add_client_accept(self=%p, listen=%p, proto=%i) = ?", self, listen, proto);

 if ( self == NULL || listen == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 protoobj = __lookup_proto(self, proto);
 if ( protoobj == NULL ) {
  err = roar_error;
  if ( roar_vio_accept(&vio, listen) != -1 ) {
   roar_vio_close(&vio);
  }
  roar_error = err;
  return -1;
 }

 __ensure_client_space(self);

 for (i = 0; i < self->clients_len; i++) {
  if ( self->clients[i].status == STATUS_UNUSED ) {
   self->clients[i].vio = roar_mm_malloc(sizeof(*(self->clients[i].vio)));
   if ( self->clients[i].vio == NULL )
    return -1;
   if ( roar_vio_accept(self->clients[i].vio, listen) == -1 ) {
    err = roar_error;
    roar_mm_free(self->clients[i].vio);
    roar_err_set(err);
    return -1;
   }
   self->clients[i].vio->flags |= ROAR_VIO_FLAGS_FREESELF;

   self->clients[i].status   = STATUS_USED;
   self->clients[i].proto    = proto;
   self->clients[i].obuf     = NULL;
   self->clients[i].userdata = NULL;
   if ( protoobj->impl->set_proto != NULL ) {
    if ( protoobj->lhandle != NULL )
     roar_dl_context_restore(protoobj->lhandle);
    protoobj->impl->set_proto(i, self->clients[i].vio, &(self->clients[i].obuf), &(self->clients[i].userdata), protoobj->para, protoobj->paralen, protoobj->pluginpara);
    if ( protoobj->lhandle != NULL )
     roar_dl_context_store(protoobj->lhandle);
   }
   ROAR_DBG("savannahice_network_add_client_accept(self=%p, listen=%p, proto=%i): id=%zu", self, listen, proto, i);
   ROAR_DBG("savannahice_network_add_client_accept(self=%p, listen=%p, proto=%i) = 0", self, listen, proto);
   return 0;
  }
 }

 ROAR_DBG("savannahice_network_add_client_accept(self=%p, listen=%p, proto=%i) = -1 // NOSPC", self, listen, proto);
 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}

int savannahice_network_add_protocol(savannahice_network_t self, const savannahice_protocol_t * proto) {
 size_t i;

 if ( self == NULL || proto == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 for (i = 0; i < MAX_PROTOS; i++) {
  if ( !self->protos[i].used ) {
   self->protos[i] = *proto;
   self->protos[i].used = 1;
   return 0;
  }
 }

 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}

int savannahice_network_check(savannahice_network_t self, int_least32_t timeout) {
 const savannahice_protocol_t * proto;
 struct roar_vio_selecttv tv;
 struct roar_vio_select * vios;
 size_t num = 0, num_clients;
 size_t i;
 ssize_t ret;
 int eventsq;
 int err;

 ROAR_DBG("savannahice_network_check(self=%p) = ?", self);

 tv.sec = 0;
 tv.nsec = timeout*1000;

 vios = roar_mm_malloc(sizeof(*vios) * (self->clients_len + MAX_LISTENS));
 // terminate in case we have no memory left for vios structure...
 if ( vios == NULL )
  return -1;

 memset(vios, 0, sizeof(vios));

 ROAR_DBG("savannahice_network_check(self=%p, timeout=%i): self->clients_len=%u", self, (int)timeout, (unsigned int)self->clients_len);
 for (i = 0; i < self->clients_len; i++) {
  if ( self->clients[i].status == STATUS_USED ) {
   proto = __lookup_proto(self, self->clients[i].proto);

   if ( proto == NULL ) {
    __delete_client(self, i);
    continue;
   }

   eventsq = 0;
   if ( proto->impl->status == NULL ) {
    eventsq |= ROAR_VIO_SELECT_READ;
   } else {
    if ( proto->lhandle != NULL )
     roar_dl_context_restore(proto->lhandle);
    if ( proto->impl->status(i, self->clients[i].vio, &(self->clients[i].obuf), &(self->clients[i].userdata), proto->para, proto->paralen, proto->pluginpara) & ROAR_DL_PROTO_STATUS_RX_READY )
     eventsq |= ROAR_VIO_SELECT_READ;
    if ( proto->lhandle != NULL )
     roar_dl_context_store(proto->lhandle);
   }

   if ( self->clients[i].obuf != NULL )
    eventsq |= ROAR_VIO_SELECT_WRITE;

   if ( eventsq == 0 )
    continue;

   ROAR_VIO_SELECT_SETVIO(&(vios[num]), self->clients[i].vio, eventsq);
   vios[num].ud.si = i;
   num++;
  }
 }

 num_clients = num;

 if ( __check_for_client_space(self) ) {
  for (i = 0; i < MAX_LISTENS; i++) {
   if ( self->listens[i].status == STATUS_USED ) {
    ROAR_VIO_SELECT_SETVIO(&(vios[num]), &(self->listens[i].vio), ROAR_VIO_SELECT_READ);
    vios[num].ud.si = i;
    num++;
   }
  }
 }

 // auto terminate when no clients are connected any longer.
 if ( num == 0 ) {
  err = roar_error;
  roar_mm_free(vios);
  roar_err_set(err);
  return -1;
 }

 ROAR_DBG("savannahice_network_check(self=%p) = ?", self);
 ret = roar_vio_select(vios, num, &tv, NULL);
 if ( ret == 0 ) {
  roar_mm_free(vios);
  return 0;
 } if ( ret < 0 ) {
  err = roar_error;
  roar_mm_free(vios);
  roar_err_set(err);
  return -1;
 }

 for (i = 0; i < num; i++) {
  if ( vios[i].eventsa == 0 )
   continue;

  if ( i < num_clients ) {
   // is client.
   proto = __lookup_proto(self, self->clients[vios[i].ud.si].proto);

   if ( proto == NULL ) {
    __delete_client(self, vios[i].ud.si);
    continue;
   }

   if ( vios[i].eventsa & ROAR_VIO_SELECT_WRITE ) {
    if ( proto->impl->flush != NULL ) {
     if ( proto->lhandle != NULL )
      roar_dl_context_restore(proto->lhandle);
     proto->impl->flush(vios[i].ud.si, vios[i].vio, &(self->clients[vios[i].ud.si].obuf), &(self->clients[vios[i].ud.si].userdata), proto->para, proto->paralen, proto->pluginpara);
     if ( proto->lhandle != NULL )
      roar_dl_context_store(proto->lhandle);
    } else {
     __flush_client(self, vios[i].ud.si, vios[i].vio, &(self->clients[vios[i].ud.si].obuf));
    }

    if ( self->clients[vios[i].ud.si].obuf == NULL && proto->impl->flushed != NULL ) {
     if ( proto->lhandle != NULL )
      roar_dl_context_restore(proto->lhandle);
     proto->impl->flushed(vios[i].ud.si, vios[i].vio, &(self->clients[vios[i].ud.si].obuf), &(self->clients[vios[i].ud.si].userdata), proto->para, proto->paralen, proto->pluginpara);
     if ( proto->lhandle != NULL )
      roar_dl_context_store(proto->lhandle);
    }
   }

   if ( vios[i].eventsa & ROAR_VIO_SELECT_READ ) {
    if ( proto->impl->handle != NULL ) {
     if ( proto->lhandle != NULL )
      roar_dl_context_restore(proto->lhandle);
     if ( proto->impl->handle(vios[i].ud.si, vios[i].vio, &(self->clients[vios[i].ud.si].obuf), &(self->clients[vios[i].ud.si].userdata), proto->para, proto->paralen, proto->pluginpara) == -1 ) {
      __delete_client(self, vios[i].ud.si);
     }
     if ( proto->lhandle != NULL )
      roar_dl_context_store(proto->lhandle);
    }
   }
  } else {
   // is listen socket.
   savannahice_network_add_client_accept(self, vios[i].vio, self->listens[vios[i].ud.si].proto);
  }
 }

 roar_mm_free(vios);
 ROAR_DBG("savannahice_network_check(self=%p) = 0", self);
 return 0;
}

int savannahice_network_client_get_userdata(savannahice_network_t self, int clientid, void *** userdata) {
 if ( self == NULL ) {
  roar_err_set(ROAR_ERROR_FAULT);
  return -1;
 }

 if ( clientid < 0 || (size_t)clientid >= self->clients_len ) {
  roar_err_set(ROAR_ERROR_RANGE);
  return -1;
 }

 *userdata = &(self->clients[clientid].userdata);

 return 0;
}

//ll
