//pool.c:

/*
 *      Copyright (C) Philipp 'ph3-der-loewe' Schafft - 2012-2013
 *
 *  This file is part of SavannahIce,
 *  a protocol converter between RoarAudio and Icecast.
 *  See README for details.
 *
 *  This file is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 3
 *  as published by the Free Software Foundation.
 *
 *  RoarAudio is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 51 Franklin Street, Fifth Floor,
 *  Boston, MA 02110-1301, USA.
 *
 */

#include "savannahice.h"

struct savannahice_pool {
 size_t refc;
 size_t len;
 size_t used;
 uint32_t * pool;
};

#define _ckpool if ( _p == NULL || (pool = *_p) == NULL ) { roar_err_set(ROAR_ERROR_FAULT); return -1; }

static void __pool_update_size(savannahice_pool_t * _p) {
 savannahice_pool_t pool = *_p;
 void * p;
 size_t size;

 ROAR_DBG("__pool_update_size(pool=%p{%p{.used=%u, .len=%u, ...}}) = ?", _p, pool, (unsigned int)pool->used, (unsigned int)pool->len);

 if ( pool->used == pool->len ) {
  ROAR_DBG("__pool_update_size(_p=%p{%p}) = ?", _p, pool);
  // increase size.
  size = pool->len + 64;
  size = ((size + 31)/32)*4;
  p = roar_mm_realloc(pool, sizeof(*pool) + size);
  if ( p == NULL )
   return;
  *_p = pool = p;
  pool->pool = p + sizeof(*pool);

  ROAR_DBG("__pool_update_size(_p=%p{%p{.pool=%p, .len=%u, ...}}): sizeof(*pool) = %u", _p, pool, pool->pool, (unsigned int)pool->len, (unsigned int)sizeof(*pool));
  ROAR_DBG("__pool_update_size(...): CALL memset(%p, 0, %u);", ((void*)pool->pool) + (pool->len/8), (unsigned int)(size*4 - pool->len)/8);
  memset(((void*)pool->pool) + pool->len/8, 0, (size*8 - pool->len)/8);

  pool->len = size*8;
  ROAR_DBG("__pool_update_size(pool=%p{.used=%u, .len=%u, ...) = ?", pool, (unsigned int)pool->used, (unsigned int)pool->len);
 } else if ( pool->used < pool->len ) {
  // reduce size.
  ROAR_DBG("__pool_update_size(pool=%p) = ?", pool);
 }

 ROAR_DBG("__pool_update_size(pool=%p{%p{.used=%u, .len=%u, ...}}) = (void)", _p, pool, (unsigned int)pool->used, (unsigned int)pool->len);
}

savannahice_pool_t savannahice_pool_new(ssize_t size) {
 savannahice_pool_t pool;
 void * p;

 if ( size < 1 ) {
  size = 32;
 }

 size = ((size + 31)/32)*4;

 p = roar_mm_malloc(sizeof(*pool) + size);

 if ( p == NULL )
  return NULL;

 pool = p;
 memset(pool, 0, sizeof(*pool) + size);
 pool->refc = 1;
 pool->len  = size*8;
 pool->used = 0;
 pool->pool = p + sizeof(*pool);

 return pool;
}

int savannahice_pool_ref(savannahice_pool_t * _p) {
 savannahice_pool_t pool;
 _ckpool
 pool->refc++;
 return 0;
}

int savannahice_pool_unref(savannahice_pool_t * _p) {
 savannahice_pool_t pool;
 _ckpool
 pool->refc--;

 if ( pool->refc != 0 ) {
  roar_mm_free(pool);
 }

 *_p = NULL;

 return 0;
}

int savannahice_pool_id_new(savannahice_pool_t * _p) {
 savannahice_pool_t pool;
 int id;
 size_t i;

 ROAR_DBG("savannahice_pool_id_new(_p=%p) = ?", _p);

 _ckpool

 if ( pool->used == pool->len ) {
  __pool_update_size(_p);
  pool = *_p;
 }

 if ( pool->used == pool->len ) {
  roar_err_set(ROAR_ERROR_NOSPC);
  return -1;
 }

 ROAR_DBG("savannahice_pool_id_new(_p=%p{%p{.len=%u, ...}}) = ?", _p, pool, (unsigned int)pool->len);

 for (i = 0; i < (pool->len/32); i++) {
  ROAR_DBG("savannahice_pool_id_new(_p=%p{%p}): i=%u", _p, pool, (unsigned int)i);
  for (id = 0; id < 32; id++) {
   if ( (pool->pool[i] & (1 << id)) == 0 ) {
    pool->pool[i] |= 1 << id;
    pool->used++;
    ROAR_DBG("savannahice_pool_id_new(_p=%p) = %i", _p, (int)(id + i*32));
    return id + i*32;
   }
  }
 }

 roar_panic(ROAR_FATAL_ERROR_UNKNOWN, "counter<->matrix missmatch");
 roar_err_set(ROAR_ERROR_NOSPC);
 return -1;
}

int savannahice_pool_id_free(savannahice_pool_t * _p, int id) {
 savannahice_pool_t pool;
 uint32_t bit;
 size_t index;

 _ckpool

 if ( id < 0 || (size_t)id >= pool->len ) {
  roar_err_set(ROAR_ERROR_INVAL);
  return -1;
 }

 index = id / 32;
 id   &= 0x1F;
 bit   = 1 << id;

 pool->pool[index] |= bit;
 pool->pool[index] -= bit;
 pool->used--;

 __pool_update_size(_p);

 return 0;
}

//ll
